from .csv import CsvFile
from .docx import DocxFile
from .pdf import PdfFile


class FileFactory:

	@staticmethod
	def create_pdf_file() -> PdfFile:
		return PdfFile()

	@staticmethod
	def create_csv_file() -> CsvFile:
		return CsvFile()

	@staticmethod
	def create_docx_file() -> DocxFile:
		return DocxFile()



"""
CORE CONSTANTS

"""
import os
from pathlib import Path


class MsbConfigNames:
	MSB_JWT_VERIFY_OWNER = "MSB_JWT_VERIFY_OWNER"
	MSB_JWT_TRUSTED_OWNERS = "MSB_JWT_TRUSTED_OWNERS"
	MSB_JWT_VERIFY_SUBSCRIPTIONS = "MSB_JWT_VERIFY_SUBSCRIPTIONS"
	MSB_JWT_VERIFY_ENVIRONMENT = "MSB_JWT_VERIFY_ENVIRONMENT"
	MSB_DB_ENCRYPT_PRIVATE_FIELDS = "MSB_DB_ENCRYPT_PRIVATE_FIELDS"
	MSB_DB_PK_IS_PRIVATE_FIELD = "MSB_DB_PK_IS_PRIVATE_FIELD"
	MSB_ISR_TOKEN = "MSB_ISR_TOKEN"

	API_GATEWAY_URL = 'API_GATEWAY_URL'
	USER_SERVICE_URL = 'USER_SERVICE_URL'
	EMPLOYEE_SERVICE_URL = 'EMPLOYEE_SERVICE_URL'
	LEAVE_SERVICE_URL = 'LEAVE_SERVICE_URL'
	PROJECT_SERVICE_URL = 'PROJECT_SERVICE_URL'
	CONF_SERVICE_URL = 'CONF_SERVICE_URL'
	NOTIFICATION_SERVICE_URL = 'NOTIFICATION_SERVICE_URL'
	AUTOPILOT_SERVICE_URL = 'AUTOPILOT_SERVICE_URL'
	MSB_DEFAULT_BRODCAST_QUEUES = 'MSB_DEFAULT_BRODCAST_QUEUES'


class EnvDefaults:
	MSB_JWT_VERIFY_OWNER = False
	MSB_JWT_TRUSTED_OWNERS = []
	MSB_DB_ENCRYPT_PRIVATE_FIELDS = False
	MSB_DB_ENCRYPT_PROTECTED_FIELDS = False
	MSB_DB_PK_IS_PRIVATE_FIELD = False
	MSB_JWT_VERIFY_SUBSCRIPTIONS = False
	MSB_JWT_VERIFY_ENVIRONMENT = False

	# allow admin module
	ALLOW_ADMIN_URL = False

	# Secure csrf cookie
	CSRF_COOKIE_SECURE = True

	# Secure session cookie
	SESSION_COOKIE_SECURE = True

	# make csrf cokkied http only
	CSRF_COOKIE_HTTPONLY = True

	# make csrf cokkie http only
	SESSION_COOKIE_HTTPONLY = True

	# secure the ssl redirect
	SECURE_SSL_REDIRECT = True

	CORS_ALLOW_METHODS = ["DELETE", "GET", "OPTIONS", "POST", "PUT", ]

	CORS_ALLOW_HEADERS = [
		"accept", "accept-encoding", "authorization", "content-type",
		"dnt", "origin", "user-agent", "x-csrftoken",
		"x-requested-with",
	]

	CORS_PREFLIGHT_MAX_AGE = 86400

	CORS_ALLOW_CREDENTIALS = True


class NameConst:
	"""
	ENV CONSTANT NAMES
	"""
	ENV_LOAD_STATUS_KEY_NAME = 'ENV_CONFIG_LOADED'
	DOT_ENV_FILE_NAME = ".env"
	LOCAL_CONFIG_FILE_PREFIX = "local"
	ENVIRONMENT_VARIABLE_NAME = "ENVIRONMENT"
	DEBUG_VARIABLE_NAME = "DEBUG"

	"""
	CONSTANT DIRECTORY NAMES
	"""
	APP_DIR_NAME = "app"
	RESOURCES_DIR_NAME = "resources"
	WRITABLE_DIR_NAME = "writable"
	FIXTURES_DIR_NAME = "fixtures"

	DEFAULT_FIXTURE_DIR_NAME = "base"
	PROD_FIXTURE_DIR_NAME = "prod"
	TEST_FIXTURE_DIR_NAME = "test"

	RESOURCE_CONFIG_DIR_NAME = "config"
	SETUP_SCRIPTS_DIR_NAME = "setup"
	TEST_DATA_DIR_NAME = "test_data"
	LOGS_DIR_NAME = "logs"
	MIGRATIONS_DIR_NAME = "migrations"

	"""
	CONSTANT ENVIRONMENT NAMES
	"""
	PROD_ENV_NAME = "prod"
	STAGE_ENV_NAME = "stage"
	TEST_ENV_NAME = "test"
	DEV_ENV_NAME = "dev"
	LOCAL_ENV_NAME = "local"

	ENV_NAME_LIST = [LOCAL_ENV_NAME, DEV_ENV_NAME, TEST_ENV_NAME, STAGE_ENV_NAME, PROD_ENV_NAME]
	DEV_OR_TEST_ENV_NAMES_LIST = [LOCAL_ENV_NAME, DEV_ENV_NAME, TEST_ENV_NAME]
	DEV_ENV_NAMES_LIST = [LOCAL_ENV_NAME, DEV_ENV_NAME]
	"""
	CONSTANT FILE NAMES
	"""
	PROD_ENV_FILE_NAME = ".env"
	DEV_ENV_FILE_NAME = ".env.local"

	"""
	CONSTANT DATABASE NAMES
	"""
	DEFAULT_DATABASE_NAME = "default"
	LOGS_DATABASE_NAME = "logs"

	YAML_FILE_EXTENTION_NAME = "yaml"

	"""
	CONSTANT SERVICES NAMES
	"""
	API_GATEWAY_SERVICE_NAME = "API_GATEWAY"
	USER_SERVICE_NAME = 'USER_SERVICE'
	EMPLOYEE_SERVICE_NAME = 'EMPLOYEE_SERVICE'
	LEAVE_SERVICE_NAME = 'LEAVE_SERVICE'
	PROJECT_SERVICE_NAME = 'PROJECT_SERVICE'
	CONF_SERVICE_NAME = 'CONF_SERVICE'
	NOTIFICATION_SERVICE_NAME = 'NOTIFICATION_SERVICE'
	AUTOPILOT_SERVICE_NAME = 'AUTOPILOT_SERVICE'

	"""
		CONSTANT  NAMES FOR VALIDATION
	"""
	BULK_SCHEMA_NAME = 'schema_list'

	"""
		ASYNC TASK AND QUEUE NAMES
	"""
	ASYNC_TASK_NAME = 'async_task'
	EMAIL_NOTIFICATION_TASK_NAME = 'email_notification'
	YAMMER_NOTIFICATION_TASK_NAME = 'yammer_notification'
	NOTIFICATION_QUEUE_NAME = 'notification_management'
	SERVICE_NAME = 'HRIN-MSB'

	"""
		DEFAULT CELERY QUEUE NAMES
	"""
	USER_SERVICE_QUEUE_NAME = 'user_service_queue'
	EMPLOYEE_SERVICE_QUEUE_NAME = 'employee_service_queue'
	LEAVE_SERVICE_QUEUE_NAME = 'leave_service_queue'
	PROJECT_SERVICE_QUEUE_NAME = 'project_service_queue'
	CONF_SERVICE_QUEUE_NAME = 'conf_service_queue'
	NOTIFICATION_SERVICE_QUEUE_NAME = 'notification_service_queue'
	AUTOPILOT_SERVICE_QUEUE_NAME = 'autopilot_service_queue'


class PathConst:
	# path list under root directory
	BASE_DIR_PATH = Path(
		Path(__file__).__str__().replace(NameConst.RESOURCES_DIR_NAME, NameConst.APP_DIR_NAME).__str__() \
			.replace(NameConst.RESOURCES_DIR_NAME, NameConst.APP_DIR_NAME) \
			.replace('.venv', NameConst.APP_DIR_NAME)
			.split(NameConst.APP_DIR_NAME)[0]
	)
	APP_DIR_PATH = BASE_DIR_PATH.joinpath(NameConst.APP_DIR_NAME)
	WRITABLE_DIR_PATH = BASE_DIR_PATH.joinpath(NameConst.WRITABLE_DIR_NAME)
	RESOURCE_DIR_PATH = BASE_DIR_PATH.joinpath(NameConst.RESOURCES_DIR_NAME)

	PROD_ENV_FILE_PATH = BASE_DIR_PATH.joinpath(NameConst.PROD_ENV_FILE_NAME)
	DEV_ENV_FILE_PATH = BASE_DIR_PATH.joinpath(NameConst.DEV_ENV_FILE_NAME)

	# path list under writable directory
	LOGS_DIR_PATH = WRITABLE_DIR_PATH.joinpath(NameConst.LOGS_DIR_NAME)

	# path list under resources directory
	FIXTURES_DIR_PATH = RESOURCE_DIR_PATH.joinpath(NameConst.FIXTURES_DIR_NAME)
	RESOURCE_CONFIG_DIR_PATH = RESOURCE_DIR_PATH.joinpath(NameConst.RESOURCE_CONFIG_DIR_NAME)
	SETUP_SCRIPTS_DIR_PATH = RESOURCE_DIR_PATH.joinpath(NameConst.SETUP_SCRIPTS_DIR_NAME)
	TEST_DATA_DIR_PATH = RESOURCE_DIR_PATH.joinpath(NameConst.TEST_DATA_DIR_NAME)

	# path list under fixtures directory
	BASE_FIXTURES_DIR_PATH = os.path.join(FIXTURES_DIR_PATH, NameConst.DEFAULT_FIXTURE_DIR_NAME)
	PROD_FIXTURES_DIR_PATH = os.path.join(FIXTURES_DIR_PATH, NameConst.PROD_FIXTURE_DIR_NAME)
	TEST_FIXTURES_DIR_PATH = os.path.join(FIXTURES_DIR_PATH, NameConst.TEST_FIXTURE_DIR_NAME)

	# compiled constants
	SYS_PATH_LIST = [
		BASE_DIR_PATH.__str__(), APP_DIR_PATH.__str__(),
		RESOURCE_DIR_PATH.__str__(), WRITABLE_DIR_PATH.__str__()
	]

	FIXTURE_DIRS_LIST = [BASE_FIXTURES_DIR_PATH, PROD_FIXTURES_DIR_PATH, TEST_FIXTURES_DIR_PATH]


__all__ = [
	"NameConst", "PathConst", "EnvDefaults", "MsbConfigNames",
]



from rest_framework.request import Request

from msb.env.constants import NameConst
from .exceptions import (ValidationExceptions as VE)
from .schema import ValidationSchema


def parse_payload_from_request(request: Request):
	try:
		_payload = request.data
		return _payload
	except Exception as e:
		raise VE.InvalidPayloadFormat


def parse_schema_and_inp_for_bulk_validation(schema: ValidationSchema, inp, bulk=False):
	_parsed_schema = schema.parsed_schema
	_parsed_inp = inp
	if bulk:
		_parsed_schema = {NameConst.BULK_SCHEMA_NAME: {'type': 'list', 'schema': {'type': 'dict', 'schema': _parsed_schema}}}
		_parsed_inp = {NameConst.BULK_SCHEMA_NAME: inp}

	return _parsed_schema, _parsed_inp


def format_list_errors_to_string(errors: dict):
	return {k: v[0] if isinstance(v, list) and len(v) > 0 else v for k, v in errors.items()}


def formatted_validation_errors(errors: dict, bulk: bool = False):
	if isinstance(errors, dict) and len(errors.keys()) > 0:
		try:
			if bulk:
				error_list: dict = errors.get(NameConst.BULK_SCHEMA_NAME)[0]
				for error_index, errors in error_list.items():
					error_list.update({error_index: format_list_errors_to_string(errors=errors[0])})
				return error_list
			else:
				return format_list_errors_to_string(errors=errors)

		except Exception:
			raise VE.ErrorFormattingFailed
	return None


def validate_against(schema: ValidationSchema = None, inp=None, unknown=False, bulk: bool = False):
	if not isinstance(schema, ValidationSchema):
		raise VE.InvalidValidationSchema

	if not ((bulk and isinstance(inp, list)) or (not bulk and isinstance(inp, dict))):
		raise VE.InvalidValidationInputDict if not bulk else VE.InvalidValidationInputList

	errors = {}
	schema, inp = parse_schema_and_inp_for_bulk_validation(schema, inp, bulk=bulk)
	try:
		from cerberus.validator import Validator
		_validator = Validator(schema, allow_unknown=unknown)
		_validator.validate(inp)
		errors = _validator.errors
	except Exception as e:
		raise VE.ValidationFailed
	return formatted_validation_errors(errors, bulk=bulk)


def validate_inp_payload(schema: ValidationSchema, request: Request, allow_unknown: bool, bulk_inp: bool):
	"""
		validate paload data if payload rules are defined
	"""
	_payload = parse_payload_from_request(request=request)
	payload_validation_errors = validate_against(
		schema=schema, inp=_payload, unknown=allow_unknown, bulk=bulk_inp
	)
	if payload_validation_errors:
		raise VE.InvalidPayloadException(errors=payload_validation_errors)


def validate_inp_parameters(schema: ValidationSchema, params: dict, allow_unknown: bool):
	"""
		validate paramerer data if payload rules are defined
	"""
	params_validation_errors = validate_against(schema=schema, inp=params, unknown=allow_unknown)
	if params_validation_errors:
		raise VE.InvalidParamsException(errors=params_validation_errors)



from .decorators import (
	validation_schema_wrapper, api_inputs,
	validate_inp_parameters, validate_inp_payload
)
from .exceptions import *
from .rules import (DefaultRules)
from .schema import (RuleSchema, InputField, ValidationSchema)
from .utils import validate_against


class Validate:
	raw_input = validate_against
	request_data = api_inputs
	params = validate_inp_parameters
	payload = validate_inp_payload


__all__ = [
	"api_inputs", "validate_inp_parameters", "validate_inp_payload", "validate_against",
	"validation_schema_wrapper", "RuleSchema", "InputField", "ValidationSchema", "DefaultRules",
]



from datetime import timedelta

from .constants import MSB_JWT_TOKEN_VALIDATOR, DefaultJwtConfig
from .users import TokenUser


def jwt_user_auth_rule(user: TokenUser):
	return user


class DefaultJwtAuthSettings:

	def __init__(self, signing_key: str, **opt):
		self.signing_key = signing_key
		self.verifying_key = opt.get('verifying_key') or signing_key
		self.token_user_class = opt.get('token_user_class') or DefaultJwtConfig.user_class
		self.user_id_claim = opt.get('user_id_claim') or DefaultJwtConfig.userid_claim
		self.user_authentication_rule = opt.get('user_authentication_rule') or DefaultJwtConfig.auth_rule
		self.algorithm = opt.get('algorithm') or DefaultJwtConfig.algorithm_hs256
		self.auth_header_types = opt.get('auth_header_types') or DefaultJwtConfig.auth_header_types
		self.auth_header_name = opt.get('auth_header_name') or DefaultJwtConfig.auth_header_name
		self.user_id_field = opt.get('user_id_field') or DefaultJwtConfig.userid_field
		self.token_type_claim = opt.get('token_type_claim') or DefaultJwtConfig.token_type_claim
		self.jti_claim = opt.get('jti_claim') or DefaultJwtConfig.jti_claim
		self.access_token_lifetime = opt.get('access_token_lifetime') or DefaultJwtConfig.access_token_lifetime
		self.refresh_token_lifetime = opt.get('refresh_token_lifetime') or DefaultJwtConfig.refresh_token_lifetime
		self.rotate_refresh_tokens = opt.get('rotate_refresh_tokens') or DefaultJwtConfig.rotate_refresh_tokens
		self.blacklist_after_rotation = opt.get(
			'blacklist_after_rotation') or DefaultJwtConfig.blacklist_after_rotation
		self.audience = opt.get('audience') or DefaultJwtConfig.audience
		self.issuer = opt.get('issuer') or DefaultJwtConfig.audience

	@property
	def authentication_classes(self):
		return [MSB_JWT_TOKEN_VALIDATOR]

	@property
	def for_token_validation(self) -> dict:
		return {
			'TOKEN_USER_CLASS': self.token_user_class,
			'USER_ID_CLAIM': self.user_id_claim,
			'USER_AUTHENTICATION_RULE': self.user_authentication_rule,
			'ALGORITHM': self.algorithm,
			'AUTH_HEADER_TYPES': self.auth_header_types,
			'AUTH_HEADER_NAME': self.auth_header_name,
			'USER_ID_FIELD': self.user_id_field,
			'TOKEN_TYPE_CLAIM': self.token_type_claim,
			'JTI_CLAIM': self.jti_claim,
			'SIGNING_KEY': self.signing_key,
			'VERIFYING_KEY': self.verifying_key,
		}

	@property
	def for_authentication(self) -> dict:
		return {
			**self.for_token_validation,
			'ROTATE_REFRESH_TOKENS': self.rotate_refresh_tokens,
			'BLACKLIST_AFTER_ROTATION': self.blacklist_after_rotation,
			'ACCESS_TOKEN_LIFETIME': timedelta(minutes=self.access_token_lifetime),
			'REFRESH_TOKEN_LIFETIME': timedelta(minutes=self.refresh_token_lifetime),
			'AUDIENCE': self.audience,
			'ISSUER': self.issuer,
		}



from msb.exceptions import AppException


class FileExceptions:
	class DefaultException(AppException):
		_message = "Something went wrong"

	class PdfException(AppException):
		_message = "PDF Error!"

	class CsvException(AppException):
		_message = "Csv Error!"

	class DocxException(AppException):
		_message = "Word File Error!"

	class PdfExistsException(AppException):
		_message = "PDF File Already Exists!"

	class CsvExistsException(AppException):
		_message = "Csv File Already Exists!"

	class DocxExistsException(AppException):
		_message = "Word File Already Exists!"



from django.db.models import QuerySet

from msb.auth import (Permissions, api_permissions)
from msb.dataclasses import SearchParameter
from msb.http import (RestResponse)
from msb.services import ApiService
from msb.validation import DefaultRules
from .api_view import ApiView
from .constants import (CrudActions, CRUD_URL_PK_NAME)
from .funcs import ApiCrudRoutes


class ApiViewset(ApiView, ApiCrudRoutes):
	service_class: ApiService = None
	validation_schema_class = None
	search_parameter_class: SearchParameter = SearchParameter
	pk_name: str = CRUD_URL_PK_NAME

	def dispatch(self, request, *args, **kwargs):

		if (pk := kwargs.get(self.pk_name, None)) is not None:
			kwargs.update({self.pk_name: self.cipher.decrypt(pk), })
		return super().dispatch(request, *args, **kwargs)

	@api_permissions(Permissions.Management, override=False)
	def create(self, *args, **kwargs) -> RestResponse:
		try:
			is_bulk = not isinstance(self.payload, dict)
			self._validate_api_input(CrudActions.create, inp=self.payload, unknown=False, bulk=is_bulk)
			_create_payload = [self.payload] if not is_bulk else self.payload

			self.service_class.create(*_create_payload, user_id=self.user.id)
			return self.api_response.success()

		except Exception as e:
			return self.api_response.exception(e)

	@api_permissions(Permissions.Management, override=False)
	def retrieve(self, *args, **kwargs) -> RestResponse:
		try:
			self._validate_api_input(CrudActions.retrieve, inp=kwargs)
			_pk = kwargs.get(self.pk_name)
			result = self.service_class.retrieve(pk=_pk, silent=False)
			if type(result) not in [list, dict]:
				result = result.dict() if hasattr(result, "dict") else []
			return self.api_response.success(data=result)
		except Exception as e:
			return self.api_response.exception(e=e)

	@api_permissions(Permissions.Management, override=False)
	def list(self, *args, **kwargs) -> RestResponse:
		try:
			limit = self.params.get('limit')
			offset = self.params.get('offset')
			query_data = self.service_class.list(limit=limit, offset=offset)
			_data = []
			if isinstance(query_data, QuerySet):
				_data = [data.list_fields for data in query_data if hasattr(data, 'list_fields')]
			else:
				_data = query_data
			return self.api_response.success(data=_data)
		except Exception as e:
			return self.api_response.exception(e=e)

	@api_permissions(Permissions.Management, override=False)
	def update(self, *args, **kwargs) -> RestResponse:
		try:
			_rule = DefaultRules.pk_validation_rule(self.pk_name)
			_pk = kwargs.get(self.pk_name)
			_update_data = self.payload
			self._validate_api_input(CrudActions.update, inp=kwargs, rule=_rule)
			self._validate_api_input(CrudActions.update, inp=_update_data)
			self.service_class.update(pk=_pk, user_id=self.user.id, **_update_data)
			return self.api_response.success()
		except Exception as e:
			return self.api_response.exception(e=e)

	@api_permissions(Permissions.Management, override=False)
	def bulk_update(self, *args, **kwargs) -> RestResponse:
		try:
			_payload_list = [self.payload] if isinstance(self.payload, dict) else self.payload
			self._validate_api_input(CrudActions.update, inp=_payload_list, bulk=True)
			for _payload in _payload_list:
				if (_pk := _payload.get(self.pk_name)) is not None:
					del _payload[self.pk_name]
					self.service_class.update(pk=_pk, user_id=self.user.id, **_payload)
			return self.api_response.success()
		except Exception as e:
			return self.api_response.exception(e=e)

	@api_permissions(Permissions.Management, override=False)
	def delete(self, *args, **kwargs) -> RestResponse:
		try:
			_rule = DefaultRules.pk_validation_rule(self.pk_name)
			_pk = kwargs.get(self.pk_name)
			self._validate_api_input(CrudActions.delete, inp=kwargs, rule=_rule)
			self.service_class.delete(pk=_pk, user_id=self.user.id)
			return self.api_response.success()
		except Exception as e:
			return self.api_response.exception(e=e)

	@api_permissions(Permissions.Management, override=False)
	def bulk_delete(self, *args, **kwargs) -> RestResponse:
		try:
			_rule = DefaultRules.pk_validation_rule(self.pk_name)
			_payload = [self.payload] if isinstance(self.payload, dict) else self.payload
			self._validate_api_input(CrudActions.delete, inp=_payload, rule=_rule, bulk=True)
			for pk_details in _payload:
				_pk = pk_details.get(self.pk_name)
				self.service_class.delete(pk=_pk, user_id=self.user.id)
			return self.api_response.success()
		except Exception as e:
			return self.api_response.exception(e=e)

	@api_permissions(Permissions.Management, override=False)
	def search(self, *args, **kwargs):
		try:
			_rule = DefaultRules.search_validation_rule()
			self._validate_api_input(CrudActions.search, inp=self.payload, rule=_rule)

			_search_parameters = self.search_parameter_class(**self.payload)
			result = self.service_class.search(params=_search_parameters)
			return self.api_response.success(data=result)
		except Exception as e:
			return self.api_response.exception(e)



from celery import Celery
from django.conf import settings

from msb.env import NameConst, MsbConfigNames
from .msb_service import MsbService


class MsbQueue:

	def __init__(self, queue, **kwargs):
		self.celery = Celery()
		self.__queue = queue
		self.__service = kwargs.get('service', settings.MSB_SERVICE_NAME)

	@property
	def queue(self) -> str:
		return self.__queue

	@property
	def service(self) -> str:
		return self.__service

	def publish(self, task_name, data, **options):
		self.celery.send_task(name=task_name, queue=self.queue, args=[self.service], kwargs=data, **options)

	def __str__(self):
		return f'<{self.__class__.__name__} {self.queue} : {self.service} >'

	def __repr__(self):
		return self.__str__()


class QueueService(MsbService):

	@property
	def user_service_queue(self) -> MsbQueue:
		return MsbQueue(queue=NameConst.USER_SERVICE_QUEUE_NAME)

	@property
	def leave_service_queue(self) -> MsbQueue:
		return MsbQueue(queue=NameConst.LEAVE_SERVICE_QUEUE_NAME)

	@property
	def employee_service_queue(self) -> MsbQueue:
		return MsbQueue(queue=NameConst.EMPLOYEE_SERVICE_QUEUE_NAME)

	@property
	def project_service_queue(self) -> MsbQueue:
		return MsbQueue(queue=NameConst.PROJECT_SERVICE_QUEUE_NAME)

	@property
	def notification_service_queue(self) -> MsbQueue:
		return MsbQueue(queue=NameConst.NOTIFICATION_SERVICE_QUEUE_NAME)

	@property
	def autopilot_service_queue(self) -> MsbQueue:
		return MsbQueue(queue=NameConst.AUTOPILOT_SERVICE_QUEUE_NAME)

	@property
	def conf_service_queue(self) -> MsbQueue:
		return MsbQueue(queue=NameConst.CONF_SERVICE_QUEUE_NAME)

	def broadcast_to(self, *queues, task_name, data, **kwargs):
		for queue in queues:
			MsbQueue(queue=queue).publish(task_name=task_name, data=data, **kwargs)

	def brodcast(self, task_name, data, **kwargs):
		self.broadcast_to(
			*getattr(settings, MsbConfigNames.MSB_DEFAULT_BRODCAST_QUEUES, []),
			task_name=task_name, data=data, **kwargs
		)

	def get_queue(self, queue_name: str) -> MsbQueue:
		return MsbQueue(queue=queue_name)



import logging

from .constants import (SIMPLE_LOG_FORMAT)


# custom log handler that emits to the database
class DatabaseLogger(logging.Handler):
	__log_callback_func = None

	def __init__(self, callback_func=None, *args, **kwargs):
		logging.Handler.__init__(self, *args, **kwargs)
		self.setFormatter(logging.Formatter(SIMPLE_LOG_FORMAT))
		self.level = logging.INFO
		self.__log_callback_func = callback_func

	def emit(self, record):
		try:
			if callable(self.__log_callback_func):
				return self.__log_callback_func(**record.__dict__)
		except Exception:
			pass



import os

from .constants import (NameConst)


# Env variable class
class EnvVar(object):
	_key = None

	_value = None

	def __init__(self, key: str = '', value: str = ''):
		self._key = key
		self._value = value

	def __str__(self):
		return f"{self._key} : {self._value}"

	def as_str(self, default: str = '') -> str:
		return default if self._value is None else self._value

	def as_int(self, default: int = 0) -> int:
		return default if self._value is None else int(self._value)

	def as_float(self, default: float = 0.0) -> float:
		return default if self._value is None else float(self._value)

	def as_bool(self, default: bool = False) -> bool:
		if self._value is not None:
			return (int(self._value) > 0) if self._value.isnumeric() \
				else self._value.lower() in ['yes', 'true', '1']
		else:
			return default

	def as_list(self, sep='|', default: list = None) -> list:
		default = default if default is not None else []
		return default if self._value is None else self._value.split(sep=sep)

	def as_tuple(self, sep='|', default: tuple = None) -> tuple:
		default = default if default is not None else tuple()
		return tuple(self.as_list(sep=sep, default=list(default)))

	def as_dict(self, sep='|', kv_sep: str = '=', default: dict = None) -> dict:
		default = default if default is not None else dict()
		lst = self.as_list(sep=sep)
		if len(lst) > 0:
			return {
				str(item).strip().split(kv_sep)[0]:
					str(item).strip().split(kv_sep)[1]
				for item in lst}

	def as_path(self, default='', joinpath: str = None):
		if joinpath is not None:
			return os.path.join(joinpath, self.as_str(default=default))
		return self.as_str(default=default)


class Config:

	@staticmethod
	def load(env_path: str = '', main_file: str = '', local_file: str = ''):
		from .core import load_config
		load_config(env_path, main_file, local_file)

	@staticmethod
	def is_loaded() -> bool:
		return os.environ.get(NameConst.ENV_LOAD_STATUS_KEY_NAME) is not None

	@staticmethod
	def get(key: str = '', default=None) -> EnvVar:
		return EnvVar(key=key, value=os.environ.get(key, default=default))

	@staticmethod
	def get_settings(name: str, default=None):
		from django.conf import settings
		return getattr(settings, name, default)

	@staticmethod
	def debug():
		return Config.get(NameConst.DEBUG_VARIABLE_NAME).as_bool(default=False)

	@staticmethod
	def env_name() -> str | None:
		return Config.get(NameConst.ENVIRONMENT_VARIABLE_NAME).as_str(default=None)

	@staticmethod
	def is_local_env():
		return Config.env_name() == 'local'

	@staticmethod
	def is_dev_or_test_env() -> bool:
		return str(Config.env_name()).lower() in NameConst.DEV_OR_TEST_ENV_NAMES_LIST



import os
from pathlib import Path

import pandas as pd

from .core import FileExceptions
from .core import FileInterface, FileBase
from .core import FileMessage
from .core import GeneratedFile


class CsvFile(FileBase, FileInterface):
	content: list = list()
	_extension = ".csv"

	def __init__(self):
		self._extension = ".csv"

	def set_content(self, content: list):
		self.content = content
		return self

	def save(self, filename: str, location: str) -> GeneratedFile:
		file = GeneratedFile()
		try:
			filename = filename
			content = self.content
			filepath = os.path.join(location, filename + self._extension)

			if os.path.exists(filepath) and self.file_overwrite is False:
				raise FileExceptions.CsvException

			Path(location).mkdir(parents=True, exist_ok=True)
			dataframe = pd.DataFrame(content)
			dataframe.to_csv(filepath, index=False)
			file.path = filepath
			file.message = FileMessage.csv()
		except Exception as e:
			file.message = e
			file.error = True
		finally:
			return file



from django.urls import path, include
from msb.validation import validate_against, ValidationSchema, ValidationExceptions

from .constants import CRUD_URL_PK_NAME, CrudMethods, REQUEST_METHODS as HttpVerbs
from .exceptions import ApiViewsetExceptions


def search_action_routes(cls, search: bool):
	actions = {HttpVerbs.POST: (CrudMethods.search if search == True else CrudMethods.not_found)}
	return path("/search", cls.as_view(actions=actions))


def single_action_routes(cls, retrieve: bool = True, update: bool = True, delete: bool = True):
	actions = {
		HttpVerbs.GET: (CrudMethods.retrieve if retrieve == True else CrudMethods.not_found),
		HttpVerbs.PUT: (CrudMethods.update if update == True else CrudMethods.not_found),
		HttpVerbs.DELETE: (CrudMethods.delete if delete == True else CrudMethods.not_found),
	}
	return path(f"/<str:{(cls.pk_name or CRUD_URL_PK_NAME)}>", cls.as_view(actions=actions))


def bulk_action_routes(cls, create: bool = True, list: bool = True, bulk_update: bool = True, bulk_delete: bool = True):
	actions = {
		HttpVerbs.GET: (CrudMethods.list if list == True else CrudMethods.not_found),
		HttpVerbs.POST: (CrudMethods.create if create == True else CrudMethods.not_found),
		HttpVerbs.PUT: (CrudMethods.bulk_update if bulk_update == True else CrudMethods.not_found),
		HttpVerbs.DELETE: (CrudMethods.bulk_delete if bulk_delete == True else CrudMethods.not_found),
	}
	return path("", cls.as_view(actions=actions))


class ApiCrudRoutes:
	_api_input_is_validated: bool = False
	_api_permission_is_validated: bool = False

	@classmethod
	def crud_routes(cls, create: bool = True, retrieve: bool = True, update: bool = True, delete: bool = True,
	                list: bool = True, bulk_update: bool = True, bulk_delete: bool = True, search: bool = True):
		return include([
			search_action_routes(cls, search),
			bulk_action_routes(cls, create, list, bulk_update, bulk_delete),
			single_action_routes(cls, retrieve, update, delete),
		])

	@classmethod
	def _validate_api_input(cls, action: str, inp=None, unknown=True, bulk=False, rule=None):
		if not cls._api_input_is_validated:

			if cls.validation_schema_class is None:
				raise ApiViewsetExceptions.SchemaValidationClassNotDefined
			_validation_rule = rule if isinstance(rule, ValidationSchema) else \
				cls.validation_schema_class.get(action, default=ValidationSchema())
			validation_errors = validate_against(schema=_validation_rule, inp=inp, unknown=unknown, bulk=bulk)

			if validation_errors is not None:
				raise ValidationExceptions.InvalidPayloadException(errors=validation_errors)

	def _validate_api_payload(self, schema: ValidationSchema, unknown=False, bulk: bool = False):
		_payload = getattr(self, 'payload', ([] if bulk else {}))
		if (validation_errors := validate_against(schema=schema, inp=_payload, unknown=unknown, bulk=bulk)) is not None:
			raise ValidationExceptions.InvalidPayloadException(errors=validation_errors)



from msb.env import NameConst, PathConst

from .dataclasses import DbVendorConfig


def log_to_console(msg, format=False, sep="\n"):
	if format:
		_sep = '#' * ((len(msg) if 30 <= len(msg) <= 100 else 100) + 10)
		msg = f"\n{_sep}\n##  {msg.capitalize()}\n{_sep}\n"
	return print(msg)


def run_management_command(*args, format=True, log=True):
	from django.core.management import call_command
	if len(args) > 0:
		if log:
			log_to_console(msg=f"running command: {' '.join(args if len(args) > 0 else 'all')} ", format=format)
		call_command(*args)


def execute_db_query(db_connection, *queries):
	if db_connection and len(queries) > 0:
		with db_connection.cursor() as cursor:
			_result = []
			for query in queries:
				print(f"Executing Query = {query}\n")
				query_result = cursor.execute(query)
				if query.lstrip(' ').lower().startswith(('select', 'show')):
					_result.append(cursor.fetchall())
				else:
					_result.append(query_result)
			return _result[0] if len(_result) == 1 else _result


def drop_database_tables(db_name):
	from django.db import connections
	con = connections[db_name]
	db_config = get_django_db_vendor_config(db_connection=con)
	if db_config:
		log_to_console(f"Dropping Tables From Database[{db_name}]", format=True)
		table_list = execute_db_query(con, db_config.query_to_list_tables)
		queries_to_drop_tables = db_config.queries_to_drop_multiple_tables(table_list, )
		execute_db_query(con, *[*queries_to_drop_tables])


def get_django_db_vendor_config(db_connection) -> DbVendorConfig:
	from .constants import DJANGO_MIGRATION_DB_VENDOR_CONFIG
	if db_connection and hasattr(db_connection, 'vendor'):
		return DJANGO_MIGRATION_DB_VENDOR_CONFIG.get(getattr(db_connection, 'vendor'))
	return None


def init_django_app(settings_dir: str = NameConst.APP_DIR_NAME, *sys_pathlist, **kwargs):
	import django, sys, os
	sys.path.extend([*PathConst.SYS_PATH_LIST, *sys_pathlist])
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", f"{settings_dir}.settings")
	os.environ.setdefault("PYTHONUNBUFFERED", "1")

	django.setup()


def require_django(_func):
	def inner_func(*args, **kwargs):
		init_django_app()
		return _func(*args, **kwargs)

	return inner_func



import base64

from .asymmetric_cipher import AsymmetricCipher
from .exceptions import CipherExceptions
from .symmetric_cipher import SymmetricCipher


class Cipher:

	@staticmethod
	def get_instance(asymetric=False):
		if asymetric:
			cipher = AsymmetricCipher()
		else:
			cipher = SymmetricCipher()
		return cipher

	@staticmethod
	def encrypt(data=None, asymetric=False):
		try:
			if data is None:
				raise CipherExceptions.NoneValue
			return Cipher.get_instance().encrypt(data)
		except Exception:
			return data

	@staticmethod
	def decrypt(data=None, asymetric=False):
		try:
			if data is None:
				raise CipherExceptions.NoneValue
			return Cipher.get_instance().decrypt(data)
		except Exception:
			return data

	@staticmethod
	def decrypt_list_items(*datalist, asymetric=False):
		try:
			return [Cipher.decrypt(i) for i in datalist]
		except Exception as e:
			return datalist

	@staticmethod
	def encrypt_list_items(*datalist, asymetric=False):
		try:
			return [Cipher.encrypt(i) for i in datalist]
		except Exception as e:
			return datalist

	@staticmethod
	def encode_base64(value):
		return base64.b64encode(bytes(str(value), 'utf-8'))

	@staticmethod
	def decode_base64(value):
		return base64.b64decode(bytes(str(value), 'utf-8'))